/*
* Copyright (c) 2023 ACOAUTO Team.
* All rights reserved.
*
* Detailed license information can be found in the LICENSE file.
*
* File: stream_server_send.c .
*
* Author: Yan.chaodong <yanchaodong@acoinfo.com>
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "vsoa_parser.h"
#include "vsoa_server.h"
#include "yyjson.h"

/* The send file name. */
#define SEND_FILE "server.send"

/* The recv buffer size. */
#ifndef MAX_SEND_LEN
#define MAX_SEND_LEN    1024
#endif

#ifndef MAX_PATH
#define MAX_PATH        260
#endif

#define CHAR_BUFFER_LEN 128

static size_t get_file_size (const char* filepath)
{
    size_t  fsize;
    FILE   *fp;

    fp = fopen(filepath, "rb");
    if (fp == NULL) {
        fprintf(stderr, "Cannot open file %s.\n", filepath);
        return -1;
    }

    fseek(fp, 0, SEEK_END);
    fsize = ftell(fp);
    fclose(fp);

    return  fsize;
}

static int do_send_file(const char* filepath, int stream)
{
    int                 read_bytes;
    ssize_t             sent;
    size_t              total           = 0;
    char                buffer[MAX_SEND_LEN];
    FILE               *fp;
    int                 send_bytes;

    fp = fopen(filepath, "rb");
    if (fp == NULL) {
        fprintf(stderr, "Cannot open file %s.\n", filepath);
        goto    out;
    }

    /* Read the file and send to stream. */
    while (!feof(fp)) {
        read_bytes = fread(buffer, sizeof(char), sizeof(buffer), fp);
        if (read_bytes <= 0) {
            goto    out;
        }

        printf("Read %d bytes.\n", read_bytes);
        sent = 0;

        do {
            send_bytes = send(stream, buffer + sent, read_bytes, 0);
            if (send_bytes < 0) {
                if (errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR) {
                    fprintf(stderr, "Continue send, retry.\n");
                    continue;
                }
                fprintf(stderr, "Invalid fault, errno is %d.\n", errno);
                goto    out;
            }
            if (send_bytes == 0) {
                fprintf(stderr, "Peer closed.\n");
                goto    out;
            }
            sent += send_bytes;
        } while (sent != read_bytes);
        total += sent;
    }

    printf("Send file %s finished, size %lu Bytes.\n", filepath, total);
    return  (0);

out:
    if (fp) {
        fclose(fp);
        fp = NULL;
    }
    return  (-1);
}

/*
 * The thread function for read file and send to stream.
*/
static void *thread_read_stream (void *arg)
{
    vsoa_server_stream_t   *stream = (vsoa_server_stream_t *)arg;
    struct timeval          timeout;
    fd_set                  fds;
    int                     count;
    int                     client;
    char                    filepath[MAX_PATH];

    /* Set client stream connection timeout to 5s. */
    timeout.tv_sec  = 5;
    timeout.tv_usec = 0;

    FD_ZERO(&fds);
    FD_SET(stream->listenfd, &fds);

    count = select(stream->listenfd + 1, &fds, NULL, NULL, &timeout);
    if (count <= 0) {
        fprintf(stderr, "Stream has no client connected.\n");
        return  (NULL);
    }

    client = vsoa_server_stream_accept(stream, NULL, NULL, 0);
    if (client < 0) {
        return  (NULL);
    }

    getcwd(filepath, sizeof(filepath));
    snprintf(filepath, sizeof(filepath), "%s/%s", filepath, SEND_FILE);

    if (do_send_file(filepath, client) < 0) {
        fprintf(stderr, "Send file failed.\n");
    }

    vsoa_server_stream_close(stream);

    return  (NULL);
}

/*
 * The VSOA callback function to read file and create send thread.
*/
static void command_read(void *arg, vsoa_server_t *server, vsoa_cli_id_t cid,
                         vsoa_header_t *vsoa_hdr, vsoa_url_t *url, vsoa_payload_t *payload)
{
    uint32_t                seqno       = vsoa_parser_get_seqno(vsoa_hdr);
    pthread_t               thread;
    vsoa_server_stream_t   *stream;
    vsoa_payload_t          send;
    char                   *fileinfo;
    char                    filepath[MAX_PATH];

    stream = vsoa_server_stream_create(server);
    if (!stream) {
        fprintf(stderr, "Cannot create server stream.\n");
        vsoa_server_cli_reply(server, cid, VSOA_STATUS_NO_MEMORY, seqno, 0, NULL);
        return;
    }
    printf("vsoa_server_stream_create stream=%p.\n", stream);

    fileinfo = (char*)malloc(CHAR_BUFFER_LEN);
    if (fileinfo == NULL) {
        fprintf(stderr, "Malloc fileinfo failed, errno is %d.\n", errno);
        return;
    }
    
    getcwd(filepath, sizeof(filepath));
    snprintf(filepath, sizeof(filepath), "%s/%s", filepath, SEND_FILE);
    snprintf(fileinfo, CHAR_BUFFER_LEN, "{\"name\":\"%s\", \"size\":%lu}", SEND_FILE, get_file_size(filepath));
    
    send.data = NULL;
    send.data_len = 0;
    send.param = fileinfo;
    send.param_len = strlen(send.param);
    /* Make a client reply, include filesize and filename. */
    vsoa_server_cli_reply(server, cid, 0, seqno, stream->tunid, &send);
    
    free(fileinfo);

    /* create dedicated thread to write stream */
    pthread_create(&thread, NULL, thread_read_stream, stream);

    printf("Replied client with tunnel ID: %d.\n", stream->tunid);
}

int main (int argc, char *argv[])
{
    vsoa_server_t      *server;
    vsoa_url_t          url_read;
    uint16_t            server_port = 3001;
    struct sockaddr_in  addr;
    int                 ret;

    fd_set              fds;
    int                 max_fd;
    int                 count = 0;
    struct timespec     timeout;

    extern char       **environ;
    char              **env             = environ;
    char               *auto_key        = "VSOA_AUTO_PORT=";
    char               *vsoa_auto_port  = "\0";

    /* Check if 'VSOA_AUTO_PORT' environment variable exists. */
    while (*env) {
        if (memcmp(*env, auto_key, (size_t)strlen(auto_key)) == 0) {
            vsoa_auto_port = *env;
        }
        env++;
    }

    /* get server port from 'VSOA_AUTO_PORT' environment variable. */
    if (strlen(vsoa_auto_port) > 1) {
        const char  s[2]     = "=";
        char       *port_str = "\0";
        char       *token;

        token = strtok(vsoa_auto_port, s);
        while (token != NULL) {
            token = strtok(NULL, s);
            if (count == 0){
                port_str = token;
            }
            count++;
        }

        server_port = strtol(port_str, NULL, 10);
    }

    /* Set server listen address and port. */
    bzero(&addr, sizeof(struct sockaddr_in));
    addr.sin_family      = AF_INET;
    addr.sin_port        = htons(server_port);
    addr.sin_addr.s_addr = INADDR_ANY;

    server = vsoa_server_create("{\"name\":\"vsoa_stream_server\"}");
    if (!server) {
        printf("Cannot create VSOA stream server.\n");
        return  (-1);
    }

    url_read.url = "/read";
    url_read.url_len = strlen(url_read.url);
    vsoa_server_add_listener(server, &url_read, command_read, NULL);

    ret = vsoa_server_start(server, (struct sockaddr *)&addr, sizeof(struct sockaddr_in));
    if (!ret) {
        printf("Cannot start VSOA stream server.\n");
        vsoa_server_close(server);
        return  (-1);
    }

    printf("Started VSOA stream server.\n");

    timeout.tv_sec  = 1;
    timeout.tv_nsec = 0;

    while (1) {
        FD_ZERO(&fds);
        max_fd = vsoa_server_fds(server, &fds);

        count = pselect(max_fd + 1, &fds, NULL, NULL, &timeout, NULL);
        if (count > 0) {
            vsoa_server_input_fds(server, &fds);
        }
    }

    return  (0);
}
/*
 * end
 */
